@@ -23,6 +23,7 @@ import traceback
 from uuid import uuid4
 from dirtools import Dir, DirState, filehash
 from pathlib import Path
+import sys
 
 
 class JobStatus(Enum):
@@ -59,7 +60,7 @@ log.setLevel(logging.DEBUG)
 
 def handle_term(signum, frame):
     log.warning('Terminated')
-    raise KeyboardInterrupt()
+    sys.exit()
 
 
 signal.signal(signal.SIGINT, handle_term)
@@ -246,6 +247,7 @@ class NumerousClient:
             print(f'No execution id found - generating new: {self._execution_id}')
 
         self._complete = True
+        self._hibernating = False
         self.writers = []
 
         self.channel = self._init_channel(server=server, port=port, secure=secure, instance_id=self._instance_id)
@@ -273,16 +275,24 @@ class NumerousClient:
 
         clear_data = get_env(clear_data, 'NUMEROUS_CLEAR_DATA', default='False') == 'True'
 
+        self._hibernate_callback = self._default_hibernation_callback
+        self._hibernate_callback_args = ([], {})
+
         if clear_data:
             self.clear_timeseries_data()
 
-        self._listen = threading.Thread(target=self._listen_terminate)
+        self._listen = threading.Thread(target=self._command_listener)
         self._listen.setDaemon(True)
         self._listen.start()
 
     def complete_execution(self):
-        log.debug(f'Completing job {self._job_id} with execution {self._execution_id}')
-        self._base_client.CompleteExecution(spm_pb2.ExecutionMessage(project_id=self._project, scenario_id=self._scenario, job_id=self._job_id, execution_id=self._execution_id))
+        log.debug(f'Completing job {self._job_id} with execution {self._execution_id}. Hibernating: {self._hibernating}')
+        print("Completing job!!", self._hibernating)
+        self._base_client.CompleteExecution(spm_pb2.CompleteExecutionMessage(
+            execution=spm_pb2.ExecutionMessage(
+                project_id=self._project, scenario_id=self._scenario,
+                job_id=self._job_id, execution_id=self._execution_id
+            ), hibernate=self._hibernating))
 
     def _init_channel(self, server, port, secure=None, instance_id=None):
         secure_channel = get_env(secure, 'SECURE_CHANNEL')
@@ -290,7 +300,7 @@ class NumerousClient:
         port = get_env(port, 'NUMEROUS_API_PORT')
 
         log.info(f"Client connecting to: {server}:{port}, using secure channel: {secure_channel}")
-        print(server,':',port,' secure: ', str(secure_channel))
+        print(server, ':', port, ' secure: ', str(secure_channel))
         if str(secure_channel) == 'True':
             print('secure connect!')
             with open(f'{os.path.dirname(os.path.abspath(__file__))}/certs/server.crt', 'rb') as f:
@@ -325,8 +335,8 @@ class NumerousClient:
         )
         self._access_token = token.val
 
-    def _listen_terminate(self):
-       for channel, message in self.subscribe_messages([".".join(['COMMAND', self._project, self._scenario, self._job_id])]):
+    def _command_listener(self):
+        for channel, message in self.subscribe_messages([".".join(['COMMAND', self._project, self._scenario, self._job_id])]):
             log.debug('Received message: '+str(message)+'\n on channel: '+str(channel))
             if 'command' in message:
                 command = message['command']
@@ -335,6 +345,11 @@ class NumerousClient:
                     signal.raise_signal(signal.SIGTERM)
                     break
 
+                if command == 'hibernate':
+                    log.warning("Received Hibernate Command")
+                    self.hibernate(message='Hibernating')
+                    break
+
     def get_scenario_document(self, scenario=None):
         if scenario is None:
             scenario = self._scenario
@@ -349,7 +364,6 @@ class NumerousClient:
 
         return scenario_doc['jobs'][self._job_id]
 
-
     def get_group_document(self, group):
         group_data = self._base_client.GetGroup(spm_pb2.Group(project=self._project, group=group))
 
@@ -360,21 +374,15 @@ class NumerousClient:
 
         return json.loads(proj_data.project_document)
 
-
     def listen_scenario_document(self, scenario=None):
-        #if project is None:
-        #    project = self._project
         if scenario is None:
             scenario = self._scenario
 
-
         for doc in self._base_client.ListenScenario(spm_pb2.Scenario(project=project, scenario=scenario)):
             json_doc = doc.scenario_document
 
             yield json.loads(json_doc), doc.files
 
-
-
     def get_scenario(self, scenario=None, path='.'):
         log.debug('Get scenario')
         if scenario is None:
@@ -1028,25 +1036,57 @@ class NumerousClient:
         channel = ".".join(['EVENT', self._project, self._scenario])
 
         self._base_client.PublishSubcsriptionMessage(spm_pb2.SubscriptionMessage(channel=channel, message=json.dumps(message)), project_id=self._project, scenario_id=self._scenario)
-    #def subscribe_channels(self, channels):
-    #    self._base_client.SubscribeForUpdates(spm_pb2.)
+
     def subscribe_messages(self, channel_patterns):
         for message in self._base_client.SubscribeForUpdates(spm_pb2.Subscription(channel_patterns=channel_patterns, scenario=self._scenario)):
             yield message.channel, json.loads(message.message)
 
     def refresh_token(self, refresh_token):
-        #print('!')
         refreshed_token = self._token_manager.GetAccessToken(spm_pb2.Token(val=refresh_token))
         return refreshed_token
 
+    @staticmethod
+    def _default_hibernation_callback(*args, **kwargs) -> None:
+        """Default function to be called when job is hibernated. Does it need functionality?"""
+        print(f"Default Hibernation Callback with arguments. args: '{args}' kwargs: '{kwargs}'.")
+
+    def set_hibernation_callback(self, func, *args, **kwargs) -> None:
+        """
+        Use this function to set a callback to be used when job is hibernated.
+        :param func: function to be called
+        """
+        self._hibernate_callback = func
+        self._hibernate_callback_args = (args, kwargs)
+
+    def hibernate(self, message: str = 'Hibernating') -> None:
+        """
+        Call this function to hibernate Client. This will enable client to reload its state
+        after being restarted at a later point.
+        """
+        self._hibernating = True
+        self._complete = True
+        self.set_scenario_progress(message=message, status='hibernating')
+        signal.raise_signal(signal.SIGTERM)
+
     def close(self):
-        #close all writers
+        # Close all writers
         for w in self.writers:
             w.close()
+
+        # Close thread ofr refreshing token
         self._access_token_refresher.stop()
 
-        if self._complete:
+        # If job has been hibernated, call the set hibernate callback
+        if self._hibernating:
+            print("hibernate...")
+            self._hibernate_callback(*self._hibernate_callback_args[0], **self._hibernate_callback_args[1])
 
+        if self._complete:
+            if not self._hibernating:
+                print("setting status...")
+                self.set_scenario_progress(message='Finished', status='finished', force=True)
+
+            print("complete...")
             self.complete_execution()
 
         if hasattr(self, 'log_handler'):
@@ -1054,11 +1094,10 @@ class NumerousClient:
             self.log_handler.close()
 
         log.debug('Client closing')
-        if hasattr(self,'log_listener'):
+        if hasattr(self, 'log_listener'):
             self.log_listener.stop()
 
 
-
 @contextmanager
 def open_client(job_id=None, project=None, scenario=None, clear_data=None, no_log=False, instance_id=None, refresh_token=None) -> NumerousClient:
     numerous_client_ = NumerousClient(job_id, project, scenario, clear_data=clear_data, no_log=no_log, instance_id=instance_id, refresh_token=refresh_token)
@@ -1067,12 +1106,19 @@ def open_client(job_id=None, project=None, scenario=None, clear_data=None, no_lo
         yield numerous_client_
 
     except Exception as e:
-        #print("CLIENT FAILED WITH ERROR CODE: ", e, traceback.format_exc())
+        logging.warning(f"CLIENT FAILED WITH ERROR CODE - w: {e} | {traceback.format_exc()}")
+        log.warning(f"CLIENT FAILED WITH ERROR CODE - wl: {e} | {traceback.format_exc()}")
+        print(f"CLIENT FAILED WITH ERROR CODE: {e} | {traceback.format_exc()}")
         numerous_client_.push_scenario_error(error=traceback.format_exc())
 
     finally:
+        print("Closing!")
+        logging.warning("Closing!w")
+        log.warning("Close!wl")
+
         numerous_client_.close()
 
+
 class StaticSource:
     def __init__(self, row_data, t0=0, t_end=None, dt=3600, generate_index=True):
         self.row = row_data
@@ -1091,9 +1137,9 @@ class StaticSource:
             if self.t_end is not None and self.t>self.t_end:
                 raise StopIteration('Reached t end of '+str(self.t_end))
 
-
         return self.row
 
+
 class DynamicSource:
     def __init__(self, client, scenario, execution, spec, t0=0, te=0, timeout=None, source_type=None, subscribe=False):
         self.timeout = timeout
@@ -1186,6 +1232,7 @@ class DynamicSource:
 
         return row
 
+
 class InputManager:
     def __init__(self, sources):
         self.sources = sources
@@ -1251,6 +1298,7 @@ class RepeatedFunction:
         self._timer.cancel()
         self.is_running = False
 
+
 class Test(NumerousClient):
     def __init__(self):
         super(Test, self).__init__()
@@ -1274,7 +1322,8 @@ class NumerousAdminClient(NumerousClient):
 
     def launch_job(self, project, scenario, job):
         self._job_manager.StartJob(spm_pb2.Job(
-            project_id=project, scenario_id=scenario, job_id=job, user_id=self._user, organization_id=self._organization
+            project_id=project, scenario_id=scenario, job_id=job,
+            user_id=self._user, organization_id=self._organization
         ))
 
     def terminate_job(self, project, scenario, job):
@@ -1303,6 +1352,9 @@ class NumerousAdminClient(NumerousClient):
     def remove_data_execution(self, scenario, execution):
         self._base_client.RemoveExecutionData(spm_pb2.Scenario(scenario=scenario, execution=execution))
 
+    def delete_execution(self, exe_id):
+        self._job_manager.DeleteExecution(spm_pb2.ExecutionId(execution_id=exe_id))
+
     def update_job_by_backend(self, project, scenario, job, exe, message, status, log=None, complete=False):
         self._base_client.SetScenarioProgress(spm_pb2.ScenarioProgress(
             project=project, scenario=scenario, job_id=job,
@@ -1331,7 +1383,65 @@ class NumerousAdminClient(NumerousClient):
         for l in self._base_client.ReadExecutionLogEntries(spm_pb2.ExecutionReadLogs(execution_id=execution, start=0, end=0)):
             yield l.log_entry, datetime.fromtimestamp(l.timestamp)
 
+    def hibernate_job(self, project_id, scenario_id, job_id, organization_id, user_id):
+        self._job_manager.HibernateJob(spm_pb2.Job(
+            project_id=project_id, scenario_id=scenario_id, job_id=job_id,
+            organization_id=organization_id, user_id=user_id
+        ))
 
+    def resume_job(self, project_id, scenario_id, job_id, organization_id, user_id):
+        self._job_manager.ResumeJob(spm_pb2.Job(
+            project_id=project_id, scenario_id=scenario_id, job_id=job_id,
+            organization_id=organization_id, user_id=user_id
+        ))
+
+    def add_job_schedule(self, project_id, scenario_id, job_id, organization_id, user_id, sleep_after, sleep_for, enable_scheduling):
+        self._job_manager.AddJobSchedule(spm_pb2.JobSchedule(
+            job=spm_pb2.Job(
+                project_id=project_id, scenario_id=scenario_id, job_id=job_id,
+                organization_id=organization_id, user_id=user_id
+            ),
+            schedule=spm_pb2.Schedule(sleep_after=sleep_after, sleep_for=sleep_for, enable_scheduling=enable_scheduling)
+        ))
+
+    def update_job_schedule(self, project_id, scenario_id, job_id, organization_id, user_id, sleep_after, sleep_for, enable_scheduling):
+        self._job_manager.AddJobSchedule(spm_pb2.JobSchedule(
+            job=spm_pb2.Job(
+                project_id=project_id, scenario_id=scenario_id, job_id=job_id,
+                organization_id=organization_id, user_id=user_id
+            ),
+            schedule=spm_pb2.Schedule(sleep_after=sleep_after, sleep_for=sleep_for, enable_scheduling=enable_scheduling)
+        ))
+
+    def remove_job_schedule(self, project_id, scenario_id, job_id, organization_id, user_id):
+        self._job_manager.RemoveJobSchedule(spm_pb2.Job(
+            project_id=project_id, scenario_id=scenario_id, job_id=job_id,
+            organization_id=organization_id, user_id=user_id
+        ))
+
+    def get_job_schedule(self, project_id, scenario_id, job_id, organization_id, user_id):
+        return self._job_manager.GetJobSchedule(spm_pb2.Job(
+            project_id=project_id, scenario_id=scenario_id, job_id=job_id,
+            organization_id=organization_id, user_id=user_id
+        ))
+
+    def add_execution_schedule(self, execution_id, sleep_after, sleep_for, enable_scheduling):
+        self._job_manager.AddExecutionSchedule(spm_pb2.ExecutionAndSchedule(
+            execution=spm_pb2.ExecutionId(execution_id=execution_id),
+            schedule=spm_pb2.Schedule(sleep_after=sleep_after, sleep_for=sleep_for, enable_scheduling=enable_scheduling)
+        ))
+
+    def update_execution_schedule(self, execution_id, sleep_after, sleep_for, enable_scheduling):
+        self._job_manager.UpdateExecutionSchedule(spm_pb2.ExecutionAndSchedule(
+            execution=spm_pb2.ExecutionId(execution_id=execution_id),
+            schedule=spm_pb2.Schedule(sleep_after=sleep_after, sleep_for=sleep_for, enable_scheduling=enable_scheduling)
+        ))
+
+    def remove_execution_schedule(self, execution_id):
+        self._job_manager.RemoveExecutionSchedule(spm_pb2.ExecutionId(execution_id=execution_id))
+
+    def get_execution_schedule(self, execution_id):
+        return self._job_manager.GetExecutionSchedule(spm_pb2.ExecutionId(execution_id=execution_id))
 
     def make_commit(self, path, repository='test', branch='master', comment=None, parent_commit_id=None, dry_run=False, snapshot=True):
 
@@ -1430,8 +1540,6 @@ class NumerousAdminClient(NumerousClient):
 
         return token
 
-
-
     def enqueue_build(self, repository, commit_id, launch_after_build=False):
         resp = self._build_manager.LaunchBuild(spm_pb2.BuildRequest(repository=repository, commit_id=commit_id, launch_after_build=launch_after_build))
         return resp.build_id
@@ -1485,4 +1593,4 @@ def open_admin_client(server=None, port=None, secure=None, refresh_token=None) -
         yield numerous_client_
 
     finally:
-        numerous_client_.close()
\ No newline at end of file
+        numerous_client_.close()
